import gym
import tensorflow as tf
from gym import error, spaces, utils
from gym.utils import seeding
import numpy as np
from os import path
import collections


class GardenEnv(gym.Env):

    def __init__(self):
        ###### Set parameters of the environment MDP
        self.m = 3
        self.n = 4
        self.s_dim = self.m * self.n
        self.a_dim = 4
        self.gamma = 0.9
        
        self.rewards = -0.1 * np.ones([self.m,self.n]) # grid world rewards
        self.rewards[0,0] += 1    
        self.attack_target = {1:[0]
                             ,2:[3]
                             ,3:[2]
                             ,4:[0]
                             ,5:[3]
                             ,6:[2]
                             ,7:[0]
                             ,8:[3]
                             ,9:[2]
                             ,10:[0]
                             ,11:[0]
                             }
        self.s0 = 2
        self.target_states = list(range(1,self.s_dim))
        self.terminal_states = [0]

    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        s_axis0 = self.st %  self.m
        s_axis1 = self.st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        rt = self.rewards[stp_axis0,stp_axis1]
        self.st = stp_axis1 * self.m + stp_axis0
        self.t += 1
        if self.st ==0: done = True
        
        return self.st, rt, done, {}

    def step_sim(self, st, at):
        s_axis0 = st %  self.m
        s_axis1 = st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        return (stp_axis1 * self.m + stp_axis0)
            
    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    
    def reset(self):
        self.t = 0
        self.st = 2
        return self.st
    
    
    
class StayMoveEnv(gym.Env):

    def __init__(self):
        ###### Set parameters of the environment MDP
        self.s_dim = 2
        self.a_dim = 2
        self.gamma = 0.9
        self.rewards = np.array([[1,0],[1,0]]) # grid world rewards
        self.attack_target = {0:[1]
                             ,1:[1]
                             }
        
        self.s0 = 0
        self.target_states = [0,1]

    def step(self,at):
        self.stb = np.copy(self.st)
        if at == 0: self.st = self.st
        if at == 1: self.st = 1-self.st
        rt = self.rewards[self.st,at]
        self.t += 1
        return self.st, rt, False, {}

    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    def reset(self):
        self.t = 0
        self.st = np.random.choice(list(range(self.s_dim)))
        return self.st
    
class ChainEnv(gym.Env):

    def __init__(self,s_dim = 6, gamma = 1.0):
        ###### Set parameters of the environment MDP
        self.s_dim = s_dim
        self.a_dim = 2
        self.gamma = gamma
        
        self.rewards = -0.1 * np.ones(self.s_dim) # grid world rewards
        self.rewards[-1] += 1    
        self.attack_target = {}
        for i in range(self.s_dim-1):
            self.attack_target[i] = [1]
            
        self.s0 = self.s_dim-2
        self.target_states = list(range(self.s_dim-1))
        self.terminal_states = [self.s_dim-1]
     
        
    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        if at == 0: self.st = max(self.st - 1,0)
        if at == 1: self.st = min(self.st + 1, self.s_dim-1)
        rt = self.rewards[self.st]
        self.t += 1
        if self.st == self.s_dim-1: done = True
        return self.st, rt, done, {}
    
    def step_sim(self, st, at):
        if at == 0: return max(st - 1,0)
        if at == 1: return min(st + 1, self.s_dim-1)

    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    def reset(self):
        self.t = 0
        self.st = self.s_dim-2
        return self.st

class TenByTenEnv(gym.Env):

    def __init__(self):
        ###### Set parameters of the environment MDP
        self.m = 10
        self.n = 10
        self.s_dim = self.m * self.n
        self.a_dim = 4
        self.gamma = 0.9
        
        self.rewards = -0.1 * np.ones([self.m,self.n]) # grid world rewards
        self.rewards[0,self.n-1] += 1
        self.attack_target = {}
#         for i in range(self.s_dim-1):
#             if i != self.m*(self.n-1):
#                 self.attack_target[i] = [0]
        for i in range(self.n-1):
            self.attack_target[self.m*i] = [3]
            
        self.s0 = self.s_dim-1
        self.target_states = [self.m*i for i in range(self.n-1)]
        self.terminal_states = [self.m*(self.n-1)]
            

    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        s_axis0 = self.st %  self.m
        s_axis1 = self.st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        rt = self.rewards[stp_axis0,stp_axis1]
        self.st = stp_axis1 * self.m + stp_axis0
        self.t += 1
        if self.st ==self.m*(self.n-1): done = True
        
        return self.st, rt, done, {}
    
    def step_sim(self, st, at):
        s_axis0 = st %  self.m
        s_axis1 = st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        return stp_axis1 * self.m + stp_axis0
        
    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    
    def reset(self):
        self.t = 0
        self.st = self.s_dim-1
        return self.st
    
class Dog1(gym.Env):

    def __init__(self, gamma = 1.0):
        ###### Set parameters of the environment MDP
        self.s_dim = 3
        self.a_dim = 2
        self.gamma = gamma
        
        self.rewards = -0.1 * np.ones(self.s_dim) # grid world rewards
        self.rewards[0] += 1 
        self.rewards[2] += 1 
        self.attack_target = {}
        self.attack_target[1] = [0]
        
        self.s0 = self.s_dim-2
        self.target_states = [1]
        self.terminal_states = [0,self.s_dim-1]
     
        
    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        if at == 0: self.st = max(self.st - 1,0)
        if at == 1: self.st = min(self.st + 1,self.s_dim-1)
        rt = self.rewards[self.st]
        self.t += 1
        if self.st == 0 or self.st == self.s_dim-1: done = True
        return self.st, rt, done, {}
    
    def step_sim(self,st,at):
        if at == 0: return max(st - 1,0)
        if at == 1: return min(st + 1,self.s_dim-1)

    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    def reset(self):
        self.t = 0
        self.st = self.s_dim-2
        return self.st
    
class Dog2(gym.Env):

    def __init__(self):
        ###### Set parameters of the environment MDP
        self.m = 2
        self.n = 3
        self.s_dim = self.m * self.n
        self.a_dim = 4 # 0 up, 1 down, 2 left, 3 right
        self.gamma = 1
        
        self.rewards = -0.1 * np.ones([self.m,self.n]) # grid world rewards
        self.rewards[0,0] += 1
        self.rewards[0,self.n-1] += 1
        self.attack_target = {1:[0]
                             ,2:[2]
                             ,3:[0,2]
                             ,5:[2]
                             }
        
        self.s0 = 3
        self.target_states = [1,2,3,5]
        self.terminal_states = [0,4]

    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        s_axis0 = self.st %  self.m
        s_axis1 = self.st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        rt = self.rewards[stp_axis0,stp_axis1]
        self.st = stp_axis1 * self.m + stp_axis0
        self.t += 1
        if self.st == 0 or self.st == 4: done = True
        
        return self.st, rt, done, {}
    
    def step_sim(self, st, at):
        s_axis0 = st %  self.m
        s_axis1 = st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        return (stp_axis1 * self.m + stp_axis0)
    
    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    
    def reset(self):
        self.t = 0
        self.st = 3
        return self.st
    
class Dog3(gym.Env):

    def __init__(self):
        ###### Set parameters of the environment MDP
        self.m = 3
        self.n = 3
        self.s_dim = self.m * self.n
        self.a_dim = 4 # 0 up, 1 down, 2 left, 3 right
        self.gamma = 1
        
        self.rewards = -0.1 * np.ones([self.m,self.n]) # grid world rewards
        self.rewards[0,0] += 1
        self.rewards[0,self.n-1] += 1
        self.attack_target = {1:[0]
                             ,2:[3]
                             ,3:[2]
                             ,4:[0]
                             ,5:[0]
                             ,7:[2]
                             ,8:[2]
                             }
        
        self.s0 = 5
        self.target_states = [1,2,3,4,5,7,8]
        self.terminal_states = [0,6]

    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        s_axis0 = self.st %  self.m
        s_axis1 = self.st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        rt = self.rewards[stp_axis0,stp_axis1]
        self.st = stp_axis1 * self.m + stp_axis0
        self.t += 1
        if self.st == 0 or self.st == 6: done = True
        
        return self.st, rt, done, {}
    
    def step_sim(self, st, at):
        s_axis0 = st %  self.m
        s_axis1 = st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        return (stp_axis1 * self.m + stp_axis0)
        
    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    
    def reset(self):
        self.t = 0
        self.st = 5
        return self.st
    
if __name__ == '__main__':
    env = StayMoveEnv()
    print(env.reset())
    print(env.step(0))
    env = GardenEnv()
    print(env.reset())
    print(env.step(0))
    env = ChainEnv()
    print(env.reset())
    print(env.step(0))
    
