import math, random
import gym
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.autograd as autograd

import TD3.utils as utils


device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class Critic(nn.Module):
    def __init__(self, state_dim, action_dim):
        self.action_dim = action_dim
        super(Critic, self).__init__()
        self.l1 = nn.Linear(state_dim, 400)
        self.l2 = nn.Linear(400, 300)
        self.l3 = nn.Linear(300, action_dim)
        
    def forward(self, x):
        x1 = F.leaky_relu(self.l1(x))
        x1 = F.leaky_relu(self.l2(x1))
        x1 = self.l3(x1)
        return x1
    
    def act(self, state, epsilon = 0):
        if random.random() > epsilon:
            state   = torch.FloatTensor(state.reshape(1, -1)).to(device)
            q_value = self.forward(state)
            action  = q_value.max(1)[1].cpu().data.numpy().flatten()[0]
        else:
            action = random.randrange(self.action_dim)
        return action

    
class DQN(object):
    def __init__(self, state_dim, action_dim, learning_rate, disc_interval = 0.1):   
        self.disc_interval = disc_interval
        self.action_dim = action_dim
        self.critic = Critic(state_dim, action_dim).to(device)
        self.critic_target = Critic(state_dim, action_dim).to(device)
        self.critic_target.load_state_dict(self.critic.state_dict())
        self.critic_optimizer = torch.optim.Adam(self.critic.parameters(), lr = learning_rate)

    def select_action(self, state, epsilon = 0):
        return self.critic.act(state, epsilon)
    
    def action_projection(self, action):
        return - (self.action_dim-1) / 2 * self.disc_interval + self.disc_interval * action
    
    def update_target(self):
        self.critic_target.load_state_dict(self.critic.state_dict())
    
    def train(self, replay_buffer, iterations, batch_size=100, discount=0.99, target_freq = 100):

        for it in range(iterations):

            # Sample replay buffer 
            x, y, u, r, d = replay_buffer.sample(batch_size)
            state = torch.FloatTensor(x).to(device)
            action = torch.LongTensor(u).to(device)
            next_state = torch.FloatTensor(y).to(device)
            done = torch.FloatTensor(1 - d).to(device)
            reward = torch.FloatTensor(r).to(device)
            
            # Select action according to policy and add clipped noise 
            q_values      = self.critic(state)
            next_q_values = self.critic(next_state)
            next_q_state_values = self.critic_target(next_state)
            
            q_value       = q_values.gather(1, action.unsqueeze(1)).squeeze(1) 
            next_q_value  = next_q_state_values.gather(1, torch.max(next_q_values, 1)[1].unsqueeze(1)).squeeze(1)
            expected_q_value = reward.squeeze(1) + discount * next_q_value * done.squeeze(1)
            
            loss = F.smooth_l1_loss(q_value, expected_q_value)
            self.critic_optimizer.zero_grad()
            loss.backward()
            self.critic_optimizer.step()
            

            # Delayed policy updates
            if it % target_freq == 0:
                self.update_target()
        print("critic_loss: ", loss)
        print("critic_value: ", q_value[0])

    def save(self, filename, directory):
        torch.save(self.critic.state_dict(), '%s/%s_critic.pth' % (directory, filename))
        torch.save(self.critic_target.state_dict(), '%s/%s_critic_target.pth' % (directory, filename))


    def load(self, filename, directory):
        self.critic.load_state_dict(torch.load('%s/%s_critic.pth' % (directory, filename),map_location=device))
        self.critic_target.load_state_dict(torch.load('%s/%s_critic_target.pth' % (directory, filename),map_location=device))
        
        
        
        