import gym, math
import tensorflow as tf
from gym import error, spaces, utils, logger
from gym.utils import seeding
import numpy as np
from os import path
import collections


class GardenEnv(gym.Env):

    def __init__(self):
        ###### Set parameters of the environment MDP
        self.m = 3
        self.n = 4
        self.s_dim = self.m * self.n
        self.a_dim = 4
        self.gamma = 0.9
        
        self.rewards = -0.1 * np.ones([self.m,self.n]) # grid world rewards
        self.rewards[0,0] += 1    
        self.attack_target = {1:[0]
                             ,2:[3]
                             ,3:[2]
                             ,4:[0]
                             ,5:[3]
                             ,6:[2]
                             ,7:[0]
                             ,8:[3]
                             ,9:[2]
                             ,10:[0]
                             ,11:[0]
                             }
        self.s0 = 2
        self.target_states = list(range(1,self.s_dim))
        self.terminal_states = [0]

    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        s_axis0 = self.st %  self.m
        s_axis1 = self.st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        rt = self.rewards[stp_axis0,stp_axis1]
        self.st = stp_axis1 * self.m + stp_axis0
        self.t += 1
        if self.st ==0: done = True
        
        return self.st, rt, done, {}

    def step_sim(self, st, at):
        s_axis0 = st %  self.m
        s_axis1 = st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        return (stp_axis1 * self.m + stp_axis0)
            
    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    
    def reset(self):
        self.t = 0
        self.st = 2
        return self.st
    
    
    
class StayMoveEnv(gym.Env):

    def __init__(self):
        ###### Set parameters of the environment MDP
        self.s_dim = 2
        self.a_dim = 2
        self.gamma = 0.9
        self.rewards = np.array([[1,0],[1,0]]) # grid world rewards
        self.attack_target = {0:[1]
                             ,1:[1]
                             }
        
        self.s0 = 0
        self.target_states = [0,1]

    def step(self,at):
        self.stb = np.copy(self.st)
        if at == 0: self.st = self.st
        if at == 1: self.st = 1-self.st
        rt = self.rewards[self.st,at]
        self.t += 1
        return self.st, rt, False, {}

    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    def reset(self):
        self.t = 0
        self.st = np.random.choice(list(range(self.s_dim)))
        return self.st

    
class ChainEnv(gym.Env):

    def __init__(self,s_dim = 6, gamma = 1.0):
        ###### Set parameters of the environment MDP
        self.s_dim = s_dim
        self.a_dim = 2
        self.gamma = gamma
        
        self.rewards = -0.1 * np.ones(self.s_dim) # grid world rewards
        self.rewards[-1] += 1    
        self.attack_target = {}
        for i in range(self.s_dim-1):
            self.attack_target[i] = [1]
            
        self.s0 = self.s_dim-2
        self.target_states = list(range(self.s_dim-1))
        self.terminal_states = [self.s_dim-1]
     
        
    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        if at == 0: self.st = max(self.st - 1,0)
        if at == 1: self.st = min(self.st + 1, self.s_dim-1)
        rt = self.rewards[self.st]
        self.t += 1
        if self.st == self.s_dim-1: done = True
        return self.st, rt, done, {}
    
    def step_sim(self, st, at):
        if at == 0: return max(st - 1,0)
        if at == 1: return min(st + 1, self.s_dim-1)

    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    def reset(self):
        self.t = 0
        self.st = self.s_dim-2
        return self.st
    
    def randomize(self):
        self.t = 0
        self.st = np.random.choice(list(range(self.s_dim-1)))
        return self.st

class Twisted_ChainEnv(gym.Env):

    def __init__(self,s_dim = 6, gamma = 0.9):
        ###### Set parameters of the environment MDP
        self.s_dim = s_dim
        self.a_dim = 2
        self.gamma = gamma
        
        self.rewards = -0.1 * np.ones(self.s_dim) # grid world rewards
        self.rewards[-1] += 1    
        self.attack_target = {}
        for i in range(self.s_dim-1):
            self.attack_target[i] = [1]
        self.attack_target[0] = [0]
            
        self.s0 = self.s_dim-2
        self.target_states = list(range(self.s_dim-1))
        self.terminal_states = [self.s_dim-1]
     
        
    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        if at == 0: self.st = max(self.st - 1,0)
        if at == 1: self.st = min(self.st + 1, self.s_dim-1)
        rt = self.rewards[self.st]
        self.t += 1
        if self.st == self.s_dim-1: done = True
        return self.st, rt, done, {}
    
    def step_sim(self, st, at):
        if at == 0: return max(st - 1,0)
        if at == 1: return min(st + 1, self.s_dim-1)

    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    def reset(self):
        self.t = 0
        self.st = self.s_dim-2
        return self.st
    
    def randomize(self):
        self.t = 0
        self.st = np.random.choice(list(range(self.s_dim-1)))
        return self.st
    
    
class TenByTenEnv(gym.Env):

    def __init__(self):
        ###### Set parameters of the environment MDP
        self.m = 10
        self.n = 10
        self.s_dim = self.m * self.n
        self.a_dim = 4
        self.gamma = 0.9
        
        self.rewards = -0.1 * np.ones([self.m,self.n]) # grid world rewards
        self.rewards[0,self.n-1] += 1
        self.attack_target = {}
#         for i in range(self.s_dim-1):
#             if i != self.m*(self.n-1):
#                 self.attack_target[i] = [0]
        for i in range(self.n-1):
            self.attack_target[self.m*i] = [3]
            
        self.s0 = self.s_dim-1
        self.target_states = [self.m*i for i in range(self.n-1)]
        self.terminal_states = [self.m*(self.n-1)]
            

    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        s_axis0 = self.st %  self.m
        s_axis1 = self.st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        rt = self.rewards[stp_axis0,stp_axis1]
        self.st = stp_axis1 * self.m + stp_axis0
        self.t += 1
        if self.st ==self.m*(self.n-1): done = True
        
        return self.st, rt, done, {}
    
    def step_sim(self, st, at):
        s_axis0 = st %  self.m
        s_axis1 = st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        return stp_axis1 * self.m + stp_axis0
        
    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    
    def reset(self):
        self.t = 0
        self.st = self.s_dim-1
        return self.st

class TenByTenEnv1(gym.Env):

    def __init__(self):
        ###### Set parameters of the environment MDP
        self.m = 10
        self.n = 10
        self.s_dim = self.m * self.n
        self.a_dim = 4
        self.gamma = 0.9
        
        self.rewards = -0.1 * np.ones([self.m,self.n]) # grid world rewards
        self.rewards[0,self.n-1] += 1
        self.attack_target = {}
        for i in range(1,self.m):
            self.attack_target[i] = [0]
#         for i in range(self.n-1):
#             self.attack_target[self.m*i] = [3]
            
        self.s0 = self.s_dim-1
        self.target_states1 = [i for i in range(1,self.m)]
        self.target_states2 = [self.m*i for i in range(self.n-1)]
        self.target_states = self.target_states1# + self.target_states2
        self.terminal_states = [self.m*(self.n-1)]
            

    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        s_axis0 = self.st %  self.m
        s_axis1 = self.st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        rt = self.rewards[stp_axis0,stp_axis1]
        self.st = stp_axis1 * self.m + stp_axis0
        self.t += 1
        if self.st ==self.m*(self.n-1): done = True
        
        return self.st, rt, done, {}
    
    def step_sim(self, st, at):
        s_axis0 = st %  self.m
        s_axis1 = st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        return stp_axis1 * self.m + stp_axis0
        
    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    
    def reset(self):
        self.t = 0
        self.st = self.s_dim-1
        return self.st

class TenByTenEnv2(gym.Env):

    def __init__(self):
        ###### Set parameters of the environment MDP
        self.m = 10
        self.n = 10
        self.s_dim = self.m * self.n
        self.a_dim = 4
        self.gamma = 0.9
        
        self.rewards = -0.1 * np.ones([self.m,self.n]) # grid world rewards
        self.rewards[0,self.n-1] += 1
        self.attack_target = {}
        for i in range(1,self.m):
            self.attack_target[i] = [0]
        for i in range(self.n-1):
            self.attack_target[self.m*i] = [3]
            
        self.s0 = self.s_dim-1
        self.target_states1 = [i for i in range(1,self.m)]
        self.target_states2 = [self.m*i for i in range(self.n-1)]
        self.target_states = self.target_states1 + self.target_states2
        self.terminal_states = [self.m*(self.n-1)]
            

    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        s_axis0 = self.st %  self.m
        s_axis1 = self.st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        rt = self.rewards[stp_axis0,stp_axis1]
        self.st = stp_axis1 * self.m + stp_axis0
        self.t += 1
        if self.st ==self.m*(self.n-1): done = True
        
        return self.st, rt, done, {}
    
    def step_sim(self, st, at):
        s_axis0 = st %  self.m
        s_axis1 = st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        return stp_axis1 * self.m + stp_axis0
        
    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    
    def reset(self):
        self.t = 0
        self.st = self.s_dim-1
        return self.st
    
class TenByTenEnv3(gym.Env):

    def __init__(self):
        ###### Set parameters of the environment MDP
        self.m = 10
        self.n = 10
        self.s_dim = self.m * self.n
        self.a_dim = 4
        self.gamma = 0.9
        
        self.rewards = -0.1 * np.ones([self.m,self.n]) # grid world rewards
        self.rewards[0,self.n-1] += 1
        self.attack_target = {}
        for i in range(0,self.m-1):
            self.attack_target[i] = [1]
        for i in range(1,self.n-1):
            self.attack_target[self.m*i] = [2]
            
        self.s0 = self.s_dim-1
        self.target_states1 = [i for i in range(0,self.m-1)]
        self.target_states2 = [self.m*i for i in range(1,self.n-1)]
        self.target_states = self.target_states1 + self.target_states2
        self.terminal_states = [self.m*(self.n-1)]
            

    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        s_axis0 = self.st %  self.m
        s_axis1 = self.st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        rt = self.rewards[stp_axis0,stp_axis1]
        self.st = stp_axis1 * self.m + stp_axis0
        self.t += 1
        if self.st ==self.m*(self.n-1): done = True
        
        return self.st, rt, done, {}
    
    def step_sim(self, st, at):
        s_axis0 = st %  self.m
        s_axis1 = st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        return stp_axis1 * self.m + stp_axis0
        
    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    
    def reset(self):
        self.t = 0
        self.st = self.s_dim-1
        return self.st
    
class Dog1(gym.Env):

    def __init__(self, gamma = 1.0):
        ###### Set parameters of the environment MDP
        self.s_dim = 3
        self.a_dim = 2
        self.gamma = gamma
        
        self.rewards = -0.1 * np.ones(self.s_dim) # grid world rewards
        self.rewards[0] += 1 
        self.rewards[2] += 1 
        self.attack_target = {}
        self.attack_target[1] = [0]
        
        self.s0 = self.s_dim-2
        self.target_states = [1]
        self.terminal_states = [0,self.s_dim-1]
     
        
    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        if at == 0: self.st = max(self.st - 1,0)
        if at == 1: self.st = min(self.st + 1,self.s_dim-1)
        rt = self.rewards[self.st]
        self.t += 1
        if self.st == 0 or self.st == self.s_dim-1: done = True
        return self.st, rt, done, {}
    
    def step_sim(self,st,at):
        if at == 0: return max(st - 1,0)
        if at == 1: return min(st + 1,self.s_dim-1)

    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    def reset(self):
        self.t = 0
        self.st = self.s_dim-2
        return self.st
    
class Dog2(gym.Env):

    def __init__(self):
        ###### Set parameters of the environment MDP
        self.m = 2
        self.n = 3
        self.s_dim = self.m * self.n
        self.a_dim = 4 # 0 up, 1 down, 2 left, 3 right
        self.gamma = 1
        
        self.rewards = -0.1 * np.ones([self.m,self.n]) # grid world rewards
        self.rewards[0,0] += 1
        self.rewards[0,self.n-1] += 1
        self.attack_target = {1:[0]
                             ,2:[2]
                             ,3:[0,2]
                             ,5:[2]
                             }
        
        self.s0 = 3
        self.target_states = [1,2,3,5]
        self.terminal_states = [0,4]

    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        s_axis0 = self.st %  self.m
        s_axis1 = self.st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        rt = self.rewards[stp_axis0,stp_axis1]
        self.st = stp_axis1 * self.m + stp_axis0
        self.t += 1
        if self.st == 0 or self.st == 4: done = True
        
        return self.st, rt, done, {}
    
    def step_sim(self, st, at):
        s_axis0 = st %  self.m
        s_axis1 = st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        return (stp_axis1 * self.m + stp_axis0)
    
    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    
    def reset(self):
        self.t = 0
        self.st = 3
        return self.st
    
class Dog3(gym.Env):

    def __init__(self):
        ###### Set parameters of the environment MDP
        self.m = 3
        self.n = 3
        self.s_dim = self.m * self.n
        self.a_dim = 4 # 0 up, 1 down, 2 left, 3 right
        self.gamma = 1
        
        self.rewards = -0.1 * np.ones([self.m,self.n]) # grid world rewards
        self.rewards[0,0] += 1
        self.rewards[0,self.n-1] += 1
        self.attack_target = {1:[0]
                             ,2:[3]
                             ,3:[2]
                             ,4:[0]
                             ,5:[0]
                             ,7:[2]
                             ,8:[2]
                             }
        
        self.s0 = 5
        self.target_states = [1,2,3,4,5,7,8]
        self.terminal_states = [0,6]

    def step(self,at):
        self.stb = np.copy(self.st)
        done = False
        s_axis0 = self.st %  self.m
        s_axis1 = self.st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        rt = self.rewards[stp_axis0,stp_axis1]
        self.st = stp_axis1 * self.m + stp_axis0
        self.t += 1
        if self.st == 0 or self.st == 6: done = True
        
        return self.st, rt, done, {}
    
    def step_sim(self, st, at):
        s_axis0 = st %  self.m
        s_axis1 = st // self.m
        if at == 0:
            stp_axis0 = max(s_axis0 - 1,0)
            stp_axis1 = s_axis1
        if at == 1:
            stp_axis0 = min(s_axis0 + 1,self.m-1)
            stp_axis1 = s_axis1
        if at == 2: 
            stp_axis1 = max(s_axis1 - 1,0)
            stp_axis0 = s_axis0
        if at == 3:
            stp_axis1 = min(s_axis1 + 1,self.n-1)
            stp_axis0 = s_axis0
        return (stp_axis1 * self.m + stp_axis0)
        
    def rollback(self):
        self.st = np.copy(self.stb)
        self.t -=1
        
    
    def reset(self):
        self.t = 0
        self.st = 5
        return self.st
    
    
class MountainCarEnv(gym.Env):
    def __init__(self):
        2.09
        self.min_position = -1.7
        self.max_position = 0.6
        self.max_speed = 0.07
        self.goal_position_min = -1.6
        self.goal_position_max = 0.5
        
        self.force=0.001
        self.gravity=0.0025

        self.low = np.array([self.min_position, -self.max_speed])
        self.high = np.array([self.max_position, self.max_speed])

        self.viewer = None

        self.a_dim = 3
        self.s_dim = 2

        self.seed()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def step(self, action):
        position, velocity = self.st
        velocity += (action-1)*self.force + math.cos(3*position)*(-self.gravity)
        velocity = np.clip(velocity, -self.max_speed, self.max_speed)
        position += velocity
        position = np.clip(position, self.min_position, self.max_position)
        if (position==self.min_position and velocity<0): velocity = 0

        done_min = bool(position <= self.goal_position_min)
        done_max = bool(position >= self.goal_position_max)
        
        reward = -0.1 + 1.0 * int(done_max)
        

        self.st = (position, velocity)
        return np.array(self.st), reward, done_min or done_max, {}

    def reset(self):
        self.st = np.array([self.np_random.uniform(low=-0.6, high=-0.4), 0])
        return np.array(self.st)


class CartPoleEnv(gym.Env):
    """
    Description:
        A pole is attached by an un-actuated joint to a cart, which moves along a frictionless track. The pendulum starts upright, and the goal is to prevent it from falling over by increasing and reducing the cart's velocity.
    Source:
        This environment corresponds to the version of the cart-pole problem described by Barto, Sutton, and Anderson
    Observation: 
        Type: Box(4)
        Num	Observation                 Min         Max
        0	Cart Position             -4.8            4.8
        1	Cart Velocity             -Inf            Inf
        2	Pole Angle                 -24 deg        24 deg
        3	Pole Velocity At Tip      -Inf            Inf
        
    Actions:
        Type: Discrete(2)
        Num	Action
        0	Push cart to the left
        1	Push cart to the right
        
        Note: The amount the velocity that is reduced or increased is not fixed; it depends on the angle the pole is pointing. This is because the center of gravity of the pole increases the amount of energy needed to move the cart underneath it
    Reward:
        Reward is 1 for every step taken, including the termination step
    Starting State:
        All observations are assigned a uniform random value in [-0.05..0.05]
    Episode Termination:
        Pole Angle is more than 12 degrees
        Cart Position is more than 2.4 (center of the cart reaches the edge of the display)
        Episode length is greater than 200
        Solved Requirements
        Considered solved when the average reward is greater than or equal to 195.0 over 100 consecutive trials.
    """
    
    metadata = {
        'render.modes': ['human', 'rgb_array'],
        'video.frames_per_second' : 50
    }

    def __init__(self, reward = 1.0):
        self.gravity = 9.8
        self.masscart = 1.0
        self.masspole = 0.1
        self.total_mass = (self.masspole + self.masscart)
        self.length = 0.5 # actually half the pole's length
        self.polemass_length = (self.masspole * self.length)
        self.force_mag = 10.0
        self.tau = 0.02  # seconds between state updates
        self.kinematics_integrator = 'euler'
        self.reward = reward

        # Angle at which to fail the episode
        self.theta_threshold_radians = 12 * 2 * math.pi / 360
        self.x_threshold = 2.4

        # Angle limit set to 2 * theta_threshold_radians so failing observation is still within bounds
        high = np.array([
            self.x_threshold * 2,
            np.finfo(np.float32).max,
            self.theta_threshold_radians * 2,
            np.finfo(np.float32).max])

        self.action_space = spaces.Discrete(2)
        self.observation_space = spaces.Box(-high, high, dtype=np.float32)
        self.s_dim = 4
        self.a_dim = 2
        self.seed()
        self.viewer = None
        self.state = None

        self.steps_beyond_done = None

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def step(self, action):
        assert self.action_space.contains(action), "%r (%s) invalid"%(action, type(action))
        state = self.state
        x, x_dot, theta, theta_dot = state
        force = self.force_mag if action==1 else -self.force_mag
        costheta = math.cos(theta)
        sintheta = math.sin(theta)
        temp = (force + self.polemass_length * theta_dot * theta_dot * sintheta) / self.total_mass
        thetaacc = (self.gravity * sintheta - costheta* temp) / (self.length * (4.0/3.0 - self.masspole * costheta * costheta / self.total_mass))
        xacc  = temp - self.polemass_length * thetaacc * costheta / self.total_mass
        if self.kinematics_integrator == 'euler':
            x  = x + self.tau * x_dot
            x_dot = x_dot + self.tau * xacc
            theta = theta + self.tau * theta_dot
            theta_dot = theta_dot + self.tau * thetaacc
        else: # semi-implicit euler
            x_dot = x_dot + self.tau * xacc
            x  = x + self.tau * x_dot
            theta_dot = theta_dot + self.tau * thetaacc
            theta = theta + self.tau * theta_dot
        self.state = (x,x_dot,theta,theta_dot)
        done =  x < -self.x_threshold \
                or x > self.x_threshold \
                or theta < -self.theta_threshold_radians \
                or theta > self.theta_threshold_radians
        done = bool(done)

        if not done:
            reward = self.reward
        elif self.steps_beyond_done is None:
            # Pole just fell!
            self.steps_beyond_done = 0
            reward = self.reward
        else:
            if self.steps_beyond_done == 0:
                logger.warn("You are calling 'step()' even though this environment has already returned done = True. You should always call 'reset()' once you receive 'done = True' -- any further steps are undefined behavior.")
            self.steps_beyond_done += 1
            reward = 0.0

        return np.array(self.state), reward, done, {}

    def reset(self):
        self.state = self.np_random.uniform(low=-0.05, high=0.05, size=(4,))
        self.steps_beyond_done = None
        return np.array(self.state)

    def render(self, mode='human'):
        screen_width = 600
        screen_height = 400

        world_width = self.x_threshold*2
        scale = screen_width/world_width
        carty = 100 # TOP OF CART
        polewidth = 10.0
        polelen = scale * (2 * self.length)
        cartwidth = 50.0
        cartheight = 30.0

        if self.viewer is None:
            from gym.envs.classic_control import rendering
            self.viewer = rendering.Viewer(screen_width, screen_height)
            l,r,t,b = -cartwidth/2, cartwidth/2, cartheight/2, -cartheight/2
            axleoffset =cartheight/4.0
            cart = rendering.FilledPolygon([(l,b), (l,t), (r,t), (r,b)])
            self.carttrans = rendering.Transform()
            cart.add_attr(self.carttrans)
            self.viewer.add_geom(cart)
            l,r,t,b = -polewidth/2,polewidth/2,polelen-polewidth/2,-polewidth/2
            pole = rendering.FilledPolygon([(l,b), (l,t), (r,t), (r,b)])
            pole.set_color(.8,.6,.4)
            self.poletrans = rendering.Transform(translation=(0, axleoffset))
            pole.add_attr(self.poletrans)
            pole.add_attr(self.carttrans)
            self.viewer.add_geom(pole)
            self.axle = rendering.make_circle(polewidth/2)
            self.axle.add_attr(self.poletrans)
            self.axle.add_attr(self.carttrans)
            self.axle.set_color(.5,.5,.8)
            self.viewer.add_geom(self.axle)
            self.track = rendering.Line((0,carty), (screen_width,carty))
            self.track.set_color(0,0,0)
            self.viewer.add_geom(self.track)

            self._pole_geom = pole

        if self.state is None: return None

        # Edit the pole polygon vertex
        pole = self._pole_geom
        l,r,t,b = -polewidth/2,polewidth/2,polelen-polewidth/2,-polewidth/2
        pole.v = [(l,b), (l,t), (r,t), (r,b)]

        x = self.state
        cartx = x[0]*scale+screen_width/2.0 # MIDDLE OF CART
        self.carttrans.set_translation(cartx, carty)
        self.poletrans.set_rotation(-x[2])

        return self.viewer.render(return_rgb_array = mode=='rgb_array')

    def close(self):
        if self.viewer:
            self.viewer.close()
            self.viewer = None
            
            
            
if __name__ == '__main__':
    env = StayMoveEnv()
    print(env.reset())
    print(env.step(0))
    env = GardenEnv()
    print(env.reset())
    print(env.step(0))
    env = ChainEnv()
    print(env.reset())
    print(env.step(0))
    
