from os import path
import numpy as np
import tensorflow as tf
import math, random, time

import gym
from gym import error, spaces, utils
from gym.utils import seeding

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.autograd as autograd

import TD3.utils as utils



class Q_Learner(gym.Env):

    def __init__(self, s_dim, a_dim, gamma, T = 10):
        ###### Set parameters of the environment MDP
        self.s_dim = s_dim
        self.a_dim = a_dim
        self.alpha = 0.9 # learning rate for Q-learning update
        self.gamma = gamma # discounting factor in the environmental MDP
        self.epsilon = 0.1 # epsilon for epsilon greedy behavior policy
        self.H = 1000 # life span of learner
        self.T = T # environment reset frequency
      
    def action(self, st):
        Qt_reshape = np.reshape(self.Qt,[self.s_dim,self.a_dim])
        rand = np.random.rand()
        if rand<=self.epsilon:
            return np.random.choice(range(self.a_dim))
        else:
            return np.random.choice(np.argwhere(Qt_reshape[st] == np.amax(Qt_reshape[st])).flatten().tolist())
        
    def reset(self):
        self.t = 0
        self.h = 0
        self.Qt = 0.0 * np.random.randn(self.s_dim*self.a_dim)
        return self.Qt
    
    def randomize(self):
        self.t = 0
        self.h = 0
        self.Qt = np.random.randn(self.s_dim*self.a_dim)
        return self.Qt

    def update(self, st, at, rt, stp, done):
        self.Qtb = np.copy(self.Qt)
        self.t += 1
        self.h += 1
        Qt_reshape = np.reshape(self.Qt,[self.s_dim,self.a_dim])
        if not done:
            Qt_reshape[st,at] = (1-self.alpha) * Qt_reshape[st,at] + self.alpha * (rt + self.gamma * max(Qt_reshape[stp]))
        else:
            self.t = 0
            Qt_reshape[st,at] = (1-self.alpha) * Qt_reshape[st,at] + self.alpha * rt
        self.Qt = np.reshape(Qt_reshape,self.s_dim * self.a_dim)            
        return self.Qt
    
    def rollback(self):
        self.Qt = np.copy(self.Qtb)
        self.t -=1
        self.h -=1
        
class Q_Learner_softmax(gym.Env):

    def __init__(self, s_dim, a_dim, gamma, T = 10):
        ###### Set parameters of the environment MDP
        self.s_dim = s_dim
        self.a_dim = a_dim
        self.alpha = 0.9 # learning rate for Q-learning update
        self.gamma = gamma # discounting factor in the environmental MDP
        self.H = 1000 # life span of learner
        self.T = T # environment reset frequency
      
    def action(self, st):
        Qt_reshape = np.reshape(self.Qt,[self.s_dim,self.a_dim])
        return np.random.choice(range(self.a_dim),p = np.exp(Qt_reshape[st])/np.sum(np.exp(Qt_reshape[st])))
        
    def reset(self):
        self.t = 0
        self.h = 0
        self.Qt = 0.0 * np.random.randn(self.s_dim*self.a_dim)
        return self.Qt
    
    
    def randomize(self):
        self.t = 0
        self.h = 0
        self.Qt = 2 * np.random.randn(self.s_dim*self.a_dim) + 3
        return self.Qt

    def update(self, st, at, rt, stp, done):
        self.Qtb = np.copy(self.Qt)
        self.t += 1
        self.h += 1
        Qt_reshape = np.reshape(self.Qt,[self.s_dim,self.a_dim])
        if not done:
            Qt_reshape[st,at] = (1-self.alpha) * Qt_reshape[st,at] + self.alpha * (rt + self.gamma * max(Qt_reshape[stp]))
        else:
            self.t = 0
            Qt_reshape[st,at] = (1-self.alpha) * Qt_reshape[st,at] + self.alpha * rt
        self.Qt = np.reshape(Qt_reshape,self.s_dim * self.a_dim)            
        return self.Qt
    
    def rollback(self):
        self.Qt = np.copy(self.Qtb)
        self.t -=1
        self.h -=1
        
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class Critic(nn.Module):
    def __init__(self, state_dim, action_dim, freeze = False):
        self.action_dim = action_dim
        super(Critic, self).__init__()
        self.l1 = nn.Linear(state_dim, 256)
        self.l2 = nn.Linear(256, 128)
        self.l3 = nn.Linear(128, action_dim)
        if freeze:
            self.l1.weight.requires_grad = False
            self.l1.bias.requires_grad = False
            self.l2.weight.requires_grad = False
            self.l2.bias.requires_grad = False
            
        
    def forward(self, x):
        x1 = F.leaky_relu(self.l1(x))
        x1 = F.leaky_relu(self.l2(x1))
        x1 = self.l3(x1)
        return x1
    
    def act(self, state, epsilon = 0):
        if random.random() > epsilon:
            state   = torch.FloatTensor(state.reshape(1, -1)).to(device)
            q_value = self.forward(state)
            action  = q_value.max(1)[1].cpu().data.numpy().flatten()[0]
        else:
            action = random.randrange(self.action_dim)
        return action

    
class DQN(object):
    def __init__(self, state_dim = 4, action_dim = 2, learning_rate = 0.01, iterations = 1,
                 buffer_size = 1, epsilon = 0.5, batch_size = 1, discount = 0.95, tau = 0.01, target_freq = 1, T = 200, freeze = False, residual = False):   
        self.state_dim = state_dim
        self.action_dim = action_dim
        self.freeze = freeze
        self.critic = Critic(self.state_dim, self.action_dim, self.freeze).to(device)
        self.critic_target = Critic(self.state_dim, self.action_dim, self.freeze).to(device)
        self.critic_target.load_state_dict(self.critic.state_dict())
        self.buffer_size = buffer_size
        self.buffer = utils.ReplayBuffer(max_size=self.buffer_size)
        self.epsilon = epsilon
        self.batch_size = batch_size
        self.discount = discount
        self.target_freq = target_freq
        self.iterations = iterations
        self.learning_rate = learning_rate
        self.T = 200
        self.H = 1000 # life span of learner
        self.param_dim = np.concatenate([x.detach().cpu().numpy().flatten() for x in list(self.critic.l3.parameters())]).size
        self.tau = tau
        self.residual = residual
    def action(self, state, epsilon = 0.5):
        return self.critic.act(state, epsilon)

    def update_target(self):
        for param, target_param in zip(self.critic.parameters(), self.critic_target.parameters()):
            target_param.data.copy_(self.tau * param.data + (1 - self.tau) * target_param.data)
#         self.critic_target.load_state_dict(self.critic.state_dict())
        
    def reset(self):
        self.t = 0
        self.h = 0
        self.critic = Critic(self.state_dim, self.action_dim, self.freeze).to(device)
        self.critic_target = Critic(self.state_dim, self.action_dim, self.freeze).to(device)
        self.critic_target.load_state_dict(self.critic.state_dict())
        if self.freeze:
            self.load('cartpole_DQN','pytorch_models')
        self.critic_optimizer = torch.optim.Adam(filter(lambda p: p.requires_grad, self.critic.parameters()), lr = self.learning_rate)
        self.buffer = utils.ReplayBuffer(max_size=self.buffer_size)
        self.Qt = np.concatenate([x.detach().cpu().numpy().flatten() for x in list(self.critic.l3.parameters())])
        return self.Qt
        
    
    def train(self, prints = False):

        for it in range(self.iterations):

            # Sample replay buffer 
            x, y, u, r, d = self.buffer.sample(self.batch_size)
            state = torch.FloatTensor(x).to(device)
            action = torch.LongTensor(u).to(device)
            next_state = torch.FloatTensor(y).to(device)
            done = torch.FloatTensor(1 - d).to(device)
            reward = torch.FloatTensor(r).to(device)
            
            # Select action according to policy and add clipped noise 
            q_values      = self.critic(state)
            next_q_values = self.critic(next_state)
            if self.residual:
                next_q_state_values = self.critic(next_state)
            else:
                next_q_state_values = self.critic_target(next_state)
            
            q_value       = q_values.gather(1, action.unsqueeze(1)).squeeze(1) 
            next_q_value  = next_q_state_values.gather(1, torch.max(next_q_values, 1)[1].unsqueeze(1)).squeeze(1)
            expected_q_value = reward.squeeze(1) + self.discount * next_q_value * done.squeeze(1)
            
            loss = F.smooth_l1_loss(q_value, expected_q_value)
            self.critic_optimizer.zero_grad()
            loss.backward()
            self.critic_optimizer.step()

            # Delayed policy updates
            if it % self.target_freq == 0:
                self.update_target()
        if prints:
            print("critic_loss: ", loss)
            print("critic_value: ", q_value[0])
        
    def update(self, st, at, rt, stp, done, prints = False):
        self.t += 1
        self.h += 1
        self.buffer.add((st, stp, at, rt, float(done)))
        self.train(prints)
        
        self.Qt = np.concatenate([x.detach().cpu().numpy().flatten() for x in list(self.critic.l3.parameters())])
        return self.Qt

    def save(self, filename, directory):
        torch.save(self.critic.state_dict(), '%s/%s_critic.pth' % (directory, filename))
        torch.save(self.critic_target.state_dict(), '%s/%s_critic_target.pth' % (directory, filename))


    def load(self, filename, directory):
        self.critic.load_state_dict(torch.load('%s/%s_critic.pth' % (directory, filename),map_location=device))
        self.critic_target.load_state_dict(torch.load('%s/%s_critic_target.pth' % (directory, filename),map_location=device))
        self.Qt = np.concatenate([x.detach().cpu().numpy().flatten() for x in list(self.critic.l3.parameters())])
        
        
if __name__ == '__main__':
    env = DQN()
    print(env.reset())
    print(env.update(0,0,0,0))
    print(env.action(0))
    